@@ -1,5 +1,4 @@
 import os
-import time
 from typing import Any, Generator, Union
 
 import pytest
@@ -14,11 +13,13 @@ from okareo.model_under_test import (
     CustomModel,
     GenerationModel,
     ModelInvocation,
+    ModelUnderTest,
     OpenAIAssistantModel,
     PineconeDb,
     QdrantDB,
 )
 from okareo_api_client.api.default import (
+    delete_model_under_test_v0_models_under_test_mut_id_delete,
     delete_test_run_v0_test_runs_delete,
     find_test_data_points_v0_find_test_data_points_post,
     update_test_data_point_v0_update_test_data_point_post,
@@ -806,85 +807,123 @@ def test_delete_eval_with_checks(
     )
 
 
-def test_submit_test_openai(
+def test_register_model_versions(
     rnd: str, okareo: Okareo, single_line_scenario_set: ScenarioSetResponse
 ) -> None:
-    mut = okareo.register_model(
-        name=f"openai-ci-run-{rnd}",
-        model=GenerationModel(
-            model_id="gpt-4.1-mini",
-            temperature=0,
-            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
-            user_prompt_template=None,
-        ),
-    )
+    muts = []
+    try:
+        for i in range(1, 4):
+            mut = okareo.register_model(
+                name=f"test_register_model_versions_mut_{rnd}",
+                model=GenerationModel(
+                    system_prompt_template="\n".join(["hello world"] * i)
+                ),
+                update=True,
+            )
+            muts.append(mut)
+            assert mut.version == i
 
-    # run an async run_test
-    eval_name = f"openai-chat-run-async-{rnd}"
-    submit_resp = mut.submit_test(
-        name=eval_name,
-        scenario=single_line_scenario_set,
-        api_key=os.environ["OPENAI_API_KEY"],
-        test_run_type=TestRunType.NL_GENERATION,
-        checks=[
-            "fluency_summary",
-            "consistency_summary",
-            "relevance_summary",
-            "coherence_summary",
-        ],  # these are added by default if not specified
-    )
-    assert submit_resp.name == eval_name
+            # get the model at the version
+            model = okareo.get_model(
+                f"test_register_model_versions_mut_{rnd}", version=i
+            )
+            assert model.version == i
 
-    # wait for the async run to finish
-    # try three times with linear backoff
-    for i in range(1, 10):
-        time.sleep(3 * i)
+            # ensure that we can run an eval on the fetched model
+            eval_name = f"test_register_model_versions_mut_eval_v{i}_{rnd}"
+            eval_run = model.run_test(
+                name=eval_name,
+                scenario=single_line_scenario_set,
+                api_key=os.environ["OPENAI_API_KEY"],
+                test_run_type=TestRunType.NL_GENERATION,
+                checks=[
+                    "fluency_summary",
+                ],
+            )
+            assert eval_run.name == eval_name
 
-        # get the test run item
-        test_run = mut.get_test_run(submit_resp.id)
-        if test_run.status == "FINISHED":
-            break
-        assert test_run.status == "RUNNING"
-    assert_metrics(test_run, num_rows=1)
+        # get the model at the latest version
+        latest_model = okareo.get_model(
+            f"test_register_model_versions_mut_{rnd}", version="latest"
+        )
+        assert latest_model.version == 3
+
+    finally:
+        # Delete all versions of the model
+        delete_model_under_test_v0_models_under_test_mut_id_delete.sync_detailed(
+            client=okareo.client, api_key=API_KEY, mut_id=muts[0].mut_id
+        )
 
 
-def test_submit_batch_model_generation(
-    rnd: str, okareo: Okareo, article_clf_scenario_set: ScenarioSetResponse
+def test_register_model_version_unchanged(
+    rnd: str,
+    okareo: Okareo,
 ) -> None:
-    def generation_rules(model_input: str) -> str:
-        # simple generation rules to ensure consistent model outputs
-        out = [s.split(" ")[0] for s in model_input.split(". ")]
-        return " ".join(out)
-
-    class GenerationModel(CustomModel):
-        def invoke(self, input_value: Any) -> ModelInvocation:
-            return ModelInvocation(
-                model_prediction=generation_rules(input_value),
-                model_input=input_value,
-                model_output_metadata={"model_data": input_value},
-                tool_calls=[{"function": "function"}],
+    """Expected behavior for identical Generation model registered multiple times: version should stay at 1"""
+    try:
+        mut = None
+        for _ in range(2):
+            mut = okareo.register_model(
+                name=f"test_register_model_version_unchanged_{rnd}",
+                model=GenerationModel(
+                    system_prompt_template=f"This is a test template {rnd}"
+                ),
+                update=True,
+            )
+            assert mut.version == 1
+    finally:
+        # Delete all versions of the model
+        if mut is not None and isinstance(mut, ModelUnderTest):
+            delete_model_under_test_v0_models_under_test_mut_id_delete.sync_detailed(
+                client=okareo.client, api_key=API_KEY, mut_id=mut.mut_id
             )
 
-    mut = okareo.register_model(
-        name=f"ci-custom-nlg-batch-{rnd}",
-        model=GenerationModel(name="test_run_batch_model_generation - GenerationModel"),
-        update=True,
-    )
-    run_async_resp = mut.submit_test(
-        name=f"ci-custom-nlg-{rnd}",
-        scenario=article_clf_scenario_set,
-        test_run_type=TestRunType.NL_GENERATION,
-        checks=["latency"],
-    )
 
-    # wait for the async run to finish
-    # try three times with linear backoff
-    for i in range(1, 10):
-        time.sleep(3 * i)
+def test_register_model_versions_custom(rnd: str, okareo: Okareo) -> None:
+    """Expected behavior for CustomModel: version should stay at 1, regardless of contents of 'invoke' method"""
+    try:
 
-        # get the test run item
-        test_run = mut.get_test_run(run_async_resp.id)
-        if test_run.status == "FINISHED":
-            break
-        assert test_run.status == "RUNNING"
-    assert_metrics(test_run, num_rows=3, custom_dimensions=["latency"])
+        class FirstCustomModel(CustomModel):
+            def __init__(self, name: str) -> None:
+                super().__init__(name=name)
+
+            def invoke(self, input_value: Any) -> ModelInvocation:
+                return ModelInvocation(
+                    model_prediction="foo",
+                    model_input=input_value,
+                )
+
+        mut = okareo.register_model(
+            name=f"test_register_model_versions_mut_custom_{rnd}",
+            model=FirstCustomModel(
+                name=f"test_register_model_versions_mut_custom_{rnd}"
+            ),
+            update=True,
+        )
+        assert mut.version == 1
+
+        class SecondCustomModel(CustomModel):
+            def __init__(self, name: str) -> None:
+                super().__init__(name=name)
+
+            def invoke(self, input_value: Any) -> ModelInvocation:
+                return ModelInvocation(
+                    model_prediction=rnd,
+                    model_input=input_value,
+                )
+
+        # Register the second custom model under the same name
+        mut = okareo.register_model(
+            name=f"test_register_model_versions_mut_custom_{rnd}",
+            model=SecondCustomModel(
+                name=f"test_register_model_versions_mut_custom_{rnd}"
+            ),
+            update=True,
+        )
+        assert mut.version == 1
+
+    finally:
+        # Delete all versions of the model
+        delete_model_under_test_v0_models_under_test_mut_id_delete.sync_detailed(
+            client=okareo.client, api_key=API_KEY, mut_id=mut.mut_id
+        )
